# Note:  This is loaded into the namespace MTEST03
# variable started 0

if {![info exists started]} {variable started 0}

if {$started != 0} {
    
    variable JS ""
    variable LogFlag
    
    if {$LogFlag == 1} {InsertLog "Last Updated: [clock format [clock seconds] -format "%B %d, %Y %H:%M:%S"]"}
    
    do_click
}

if {$started == 0} {           ;#  first time load code and define local variables
    set started 1
    
    variable JS ""
    
    proc do_redisplay {} {;}
    
    proc do_testPeakTime { params } {
	
	# set name [lindex [lindex $params 1] 0]
	set sideUnderTest  [lindex [lindex $params 1] 1]
	set asicUnderTest  [lindex [lindex $params 1] 2]
	set vcall          0;#[lindex [lindex $params 1] 3] ;# [V]
	set vcalh          0;#[lindex [lindex $params 1] 4] ;# [V]
	set pulseNumber    0;#[lindex [lindex $params 1] 5] ;# [a.u.]
	set pulseWidthMin  0;#[lindex [lindex $params 1] 6] ;# [10ns]
	set pulseWidthMax  0;#[lindex [lindex $params 1] 7] ;# [10ns]
	set stepNumber     0;#[lindex [lindex $params 1] 8] ;# [a.u.]
	set pulseDelay     0;#[lindex [lindex $params 1] 9] ;# [10ns] 
	
	do_getState 
	# Start DAQ.
	after 500 set timeout 1
	do_go
	vwait timeout
	# Mandatory to enquire state
	if { [do_getState] != 1 } {return "Daq not started properly.";}
	
	# Loop over individual side.
	foreach itrSide [set ::R3B::[set ::R3B::R3B4K].SIDEList]  {
	    
	    if {$itrSide != $sideUnderTest} {continue;}
	    
	    # Loop over individual asic.
	    foreach itrAsic [set ::R3B::[set ::R3B::R3B4K].ASICList$itrSide]  {
		
		if {$asicUnderTest != "all" && $asicUnderTest != $itrAsic } {continue;}
		
		# Enable test pulse injection
		set_asicValidate $itrSide $itrAsic 1
		
		foreach itrPkTm {0 7 15} {
		    foreach itrPlsAmp {1 48 96} {
			foreach itrPlsPrd {20000 40000 60000} {
			    
			    # Set REG.4
			    set SpPt 15;# [DN]
			    set valuePkTmAndSpPt  [expr ($itrPkTm << 4) | $SpPt];
			    set registerPkTmAndSpPt 4 ;
			    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerPkTmAndSpPt] $valuePkTmAndSpPt
			    
			    # Set file name
			    set fileName  /MIDAS/SpecData/R3B/TestData/
			    append fileName [clock format [clock seconds] -format "%Y-%m-%d-"]
			    append fileName "TestPeakTime-"
			    append fileName "Side$itrSide-Asic$itrAsic-PkTm$itrPkTm-PlsAmp$itrPlsAmp-Prd$itrPlsPrd.txt"
			    
			    # Create file and set permissions.
			    set flagOpen [catch {open $fileName w} fileInput]
			    set flagPerm [catch {file attributes $fileName -permissions rwxrwxrwx} ]
			    if { $flagOpen != 0 } { InsertLog "Cannot open $fileName." }
			    if { $flagPerm != 0 } { InsertLog "Error setting permission of $fileName." }
			    
			    # Flush file header.
			    puts $fileInput "$itrSide $itrAsic $itrPkTm $itrPlsAmp $itrPlsPrd"
			    flush $fileInput
			    
			    # Set pulse amplitude.
			    set valueVcalh [expr 64+$itrPlsAmp];# [DN]
			    set valueVcall 64 ;# [DN]
			    set registerVcall 17
			    set registerVcalh 16
			    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerVcall] $valueVcall
			    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerVcalh] $valueVcalh
			    # Loop over individual channels.
			    set channelNumber 128
			    set channelGroup 4
			    for {set itrGroup 0} {$itrGroup < $channelNumber/$channelGroup} {incr itrGroup} {
				
				# To run on subset of channels.
				# if {$itrGroup !=0} { continue;}
				
				# Power on channels.
				for {set itrChannel 0} {$itrChannel < $channelGroup} {incr itrChannel} {
				    set dummyChannel [expr $itrGroup + ($itrChannel* ($channelNumber/$channelGroup) ) ]
				    set valuePlsInj [expr int( pow (2,  $dummyChannel - ( $dummyChannel/8)*8 ) ) ]
				    set registerPlsInj [expr int ( 64 + ($dummyChannel/8) )]
				    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerPlsInj] $valuePlsInj
				    
				    set valuePwr [expr 255 - int( pow (2,  $dummyChannel - ( $dummyChannel/8)*8 ) ) ]
				    set registerPwr [expr int ( 80 + ($dummyChannel/8) )]
				    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerPwr] $valuePwr
				}
				
				# Calculate step size.
				set pulseWidthMin 50;# [*10ns]
				set pulseWidthMax 1500;# [*10ns]
				set stepSize 100;# [*10ns] 
				# Make loop on pulse width.
				for {set itrStep $pulseWidthMin} {$itrStep < $pulseWidthMax} {set itrStep [expr $itrStep+$stepSize]} {
				    
				    # Set pulse params
				    set pulseNumber   512;# [a.u.]
				    set pulseWidth    [expr $itrStep];# [10ns]
				    set pulseDelay    [expr $itrPlsPrd - $pulseWidth];# [10ns]
				    InsertLog ""
				    InsertLog "PlsWdth:$pulseWidth PlsDl:$pulseDelay PlsNum:$pulseNumber"
				    
				    # Inject pulses 
				    do_loadPulseParams $pulseWidth $pulseDelay $pulseNumber
				    
				    # Zero histograms. 
				    for {set itrChannel 0} {$itrChannel < $channelGroup} {incr itrChannel} {
					set dummyChannel [expr $itrGroup + ($itrChannel* ($channelNumber/$channelGroup) ) ]
					set fileName "E.$itrSide.$itrAsic.$dummyChannel"
					# Empty the content of the histogram 
					# Check also that the histogram has been created.
					if { [ZeroSpectrum $fileName] != 0 } { 
					    return "Histograms E.$itrSide.$itrAsic.* were not created";}
				    }
				    
				    # Inject pulses 
				    do_testPulse $pulseWidth $pulseDelay $pulseNumber
				    
				    for {set itrChannel 0} {$itrChannel < $channelGroup} {incr itrChannel} {
					set dummyChannel [expr $itrGroup + ($itrChannel* ($channelNumber/$channelGroup) )  ]
					set plot  [do_analyseSpectra $itrSide $itrAsic $dummyChannel]
					set range [lindex $plot 0]
					set data  [lindex $plot 1]
					puts $fileInput "$dummyChannel $pulseWidth $range"
					puts $fileInput $data
					flush $fileInput
				    }
				    
	 			} ;# end loop over pulse width. 
				
				# Power off channels.
				for {set itrChannel 0} {$itrChannel < $channelGroup} {incr itrChannel} {
				    set dummyChannel [expr $itrGroup + ($itrChannel* ($channelNumber/$channelGroup) ) ]
				    set registerPwr [expr int ( 80 + ($dummyChannel/8) )]
				    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerPwr] 255
				    set registerPlsInj [expr int ( 64 + ($dummyChannel/8) )]
				    WriteR3BASICRegister [expr ($itrSide << 18) + ($itrAsic << 10) + $registerPlsInj] 0
				}
				
			    } ;# end of loop over channel. 
			    
			    close $fileInput
			    
			} ;# end loop over itrPlsPrd
		    } ;# end loop over itrPlsAmp
		} ;# end loop over itrPkTm
		
		# Disable test pulse injection
		set_asicValidate $itrSide $itrAsic 0 
		
	    } ;# end of loop over asic.
	} ;# end of loop over side.
	
	# Stop DAQ
	do_getState
	after 500 set timeout 1
	do_stop
	vwait timeout
	# Mandatory to enquire state
	if { [do_getState] != 0 } {return "Daq did not stop correctly";}
	
	InsertLog "Test completed"
    }
    
    proc  do_click {}  {
	variable cgidata
	variable started
	variable LogFlag
	variable ShowLogFlag
	
	set cgidata [ncgi::nvlist]
	
	if {$LogFlag == 1} {InsertLog "cgidata= ([llength $cgidata]) $cgidata"}
	
	if {[llength $cgidata] == 0} then {return}

	if {[llength $cgidata] < 2} then {
	    InsertLog "CGI data is too short: $cgidata"
	} else {
	    if {[lindex $cgidata 0] != "Widget"} then {
		InsertLog "Invalid CGI data: $cgidata"
	    } else {
		
		set w [first [lindex $cgidata 1]]
		
		switch -glob $w {
		    
		    INITIALIZE_ASICS {InsertLog "[do_powerOffFrontEnd]";}
		    INITIALIZE_DAQ {InsertLog "[do_initializeDaq $cgidata]"}
		    RESET    {set started 0}
		    TEST_PEAKTIME   { InsertLog "[do_testPeakTime $cgidata]"}
		    
		    MainLog {}
		    undefined {}
		    default {InsertLog "unexpected CGI data:  $cgidata"}
		}
	    }
        }
        do_redisplay
    }
    
}   ;# end of if {$started == 0}


