

#    experiment run control for R3B 

global MIDAS_DataAcquisitionControlServer_Verbose
set MIDAS_DataAcquisitionControlServer_Verbose 0

global Options

global R3BASICBase
set R3BASICBase 16

global NumChans  NumASICs  NumRdOUT
set NumChans 128
set NumASICs 32
set NumRDOut 8

variable SpecBase [file join $env(MIDASBASE) SpecData R3B]
variable SpecPath

proc Dump_Raw_Data {} {return "[Dump_Raw_R3B_Data]\n"}

proc Set_Options {} {
   global Options

    LoadOptions

    if {![info exists Options(Hist_MaxChans)]} {set Options(Hist_MaxChans) [expr 8*1024*1024]}

    if {![info exists Options(R3B_GroupBase)]} {set Options(R3B_GroupBase) 1}
    if {![info exists Options(Include.R3B)]}   {set Options(Include.R3B) 1}
    
    if {![info exists Options(Rate.channels)]} {set Options(Rate.channels) [expr 128*32*2]}
    if {![info exists Options(Stat.channels)]} {set Options(Stat.channels) [expr 128*32*2]}
    if {![info exists Options(Stat.shift)]}    {set Options(Stat.shift) 0}
    if {![info exists Options(Stat.offset)]}   {set Options(Stat.offset) 0}
    
    if {![info exists Options(R3B.channels)]} {set Options(R3B.channels) 4096}
    if {![info exists Options(R3B.shift)]}    {set Options(R3B.shift) 0}
    if {![info exists Options(R3B.offset)]}   {set Options(R3B.offset) 0}

    if {![info exists Options(R3B_Histogramming_Enable)]}   {set Options(R3B_Histogramming_Enable) 1}
    
    if {![info exists Options(R3B_Hist_E_Enable)]} {set Options(R3B_Hist_E_Enable) 1}
    if {![info exists Options(R3B_Hist_Hit0_Enable)]} {set Options(R3B_Hist_Hit0_Enable) 0}
    if {![info exists Options(R3B_Hist_Hit1_Enable)]} {set Options(R3B_Hist_Hit1_Enable) 1}
    if {![info exists Options(R3B_Hist_Hit01_Enable)]} {set Options(R3B_Hist_Hit01_Enable) 0}
    if {![info exists Options(R3B_Hist_DT_Enable)]} {set Options(R3B_Hist_DT_Enable) 0}

    if {![info exists Options(R3B.DTchannels)]} {set Options(R3B.DTchannels) 4000}
    if {![info exists Options(R3B.DTshift)]}    {set Options(R3B.DTshift) 4}
    if {![info exists Options(R3B.DToffset)]}   {set Options(R3B.DToffset) 0}
    
    if {![info exists Options(R3B_ValPulseWidth)]} {set Options(R3B_ValPulseWidth) 800}
    if {![info exists Options(R3B_ValPulseDelay)]} {set Options(R3B_ValPulseDelay) 19200}
    if {![info exists Options(R3B_ResetPulseWidth)]} {set Options(R3B_ResetPulseWidth) 2000}
    if {![info exists Options(R3B_Fpga_ts_reg)]} {set Options(R3B_Fpga_ts_reg) 1000}
    if {![info exists Options(R3B_Asic_ts_PulseWidth)]} {set Options(R3B_Asic_ts_PulseWidth) 8}

    for {set i 0} {$i < 2} {incr i} {
        for {set j 0} {$j < 16} {incr j} {
           if {![info exists Options(R3B_Hist_[set i].[set j]_Enable)]} {set Options(R3B_Hist_[set i].[set j]_Enable) 1}
        }
    }

    if {![info exists Options(R3B_ClockSource)]} {set Options(R3B_ClockSource) I}
    if {![info exists Options(ASIC.settings)]} {set Options(ASIC.settings) 2013}

    SaveOptions
    
}

proc Dump_Raw_R3B_Data {} {
    
    set info "data $::DataAcquisition::ACQSERVER\n"
    
    for {set offset 0} {$offset <= 500} {incr offset 100} {
	
	DataAcqCommand 19 "Getting Raw Data" "19 $offset" 2
	set data [ReadNetVar NetVar.EXEC.Response]
	
	set c 0
	
	for {set i 0} {$i < [llength $data]} {incr i} {
	    set word [lindex $data $i]
	    if {$word == 0xaa5555aa} then {
		append info "\n$word - end of dump\n"
		break
	    }
	    if {$word == 0x55aaaa55} then {
		append info "\n$word - end of event\n"
		continue
	    }
	    if {[expr ($word >> 30) & 3] == 3} {
		incr i
		set t [lindex $data $i]
		set hit [expr ($word >> 29) & 1]
		set chan [expr ($word >> 12) & 127]
		set asic [expr ($word >> 19) & 15]
		set id [expr ($word >> 23) & 0x3f]
		set adc [format "0x%x" [expr $word & 0xfff]]
		append info "\nData Item $word $t; Hit=$hit; ID=$id; ASIC=$asic; Channel=$chan; ADC=$adc; Time=$t"
	    }
	    
	    if {[expr ($word >> 30) & 3] == 2} {
	      if {[expr ($word >> 20) & 15] == 4 || [expr ($word >> 20) & 15] == 5} {
		incr i
		set t [lindex $data $i]
		set time [format "0x%x" [expr $word & 0x000fffff]]
		append info "\nWR Time Item $word $t; Time (high)=$time; Time (low)=$t"
              }
	      if {[expr ($word >> 20) & 15] == 7 || [expr ($word >> 20) & 15] == 8} {
		incr i
		set t [lindex $data $i]
		set time [format "0x%x" [expr $word & 0x000fffff]]
		append info "\nR3B Time Item $word $t; Time (high)=$time; Time (low)=$t"
              }
	      if {[expr ($word >> 20) & 15] == 14 || [expr ($word >> 20) & 15] == 15} {
		incr i
		set t [lindex $data $i]
		set time [format "0x%x" [expr $word & 0x000fffff]]
		append info "\nR3B Ext Event $word $t; Time (high)=$time; Time (low)=$t"
              }
	    }

	    if {[expr ($word >> 30) & 3] == 0} {
		append info "\nDebug Item $word"
	    }
	    
	    incr c
	    
	    if {$c == 10} then {append info "\n"; set c 0}
	}
    }
    
    return "$info"
}

proc create_R3B_spectra {} {
    global Options
    global NumChans
    global SIDEList
    global ASICList0
    global ASICList1

    if {$Options(R3B_Histogramming_Enable) == 0} {return}
    
    set Me $::DataAcquisition::ACQSERVER

    set Chs 0

puts "Hardware Config: SIDEList = $SIDEList; ASICList = $ASICList0; $ASICList1"
    
    foreach SIDE $SIDEList {
	foreach ASIC [set ASICList$SIDE] {

        if {$Options(R3B_Hist_[set SIDE].[set ASIC]_Enable) == 1} {

           if {$Chs > $Options(Hist_MaxChans)} {puts "Too many histogram channels requested ($Chs > $Options(Hist_MaxChans))"; return}
	    
	    puts "Creating spectra for SIDE $SIDE ASIC $ASIC"
	    
	    #    creating spectra for ADC data
	    
	    for {set Ch 0} {$Ch < $NumChans} {incr Ch} {
		
		if {$Options(R3B_Hist_E_Enable) == 1} {
		    
		    set z [catch {CreateSpectrum1D E.$SIDE.$ASIC.$Ch $Options(R3B.channels)} m]
		    
		    #    Allow "Spectrum already exists" so that further spectra can be created
		    if {$z!=0 && $m==0x30004} then {set z 0}
		    if {$z!=0} then {
			puts "Error attempting to create spectrum E.$SIDE.$ASIC.$Ch"
			error $m
		    }
		    
		    set z [catch {SpectrumPutTitle E.$SIDE.$ASIC.$Ch [set Me]_E.$SIDE.$ASIC.$Ch} m]

                    incr Chs $Options(R3B.channels)
		    
		}
	    }

	    #    creating spectra for time difference
	    
	    for {set Ch 0} {$Ch < $NumChans} {incr Ch} {
		
		if {$Options(R3B_Hist_DT_Enable) == 1} {
		    
		    set z [catch {CreateSpectrum1D DT.$SIDE.$ASIC.$Ch $Options(R3B.DTchannels)} m]
		    
		    #    Allow "Spectrum already exists" so that further spectra can be created
		    if {$z!=0 && $m==0x30004} then {set z 0}
		    if {$z!=0} then {
			puts "Error attempting to create spectrum DT.$SIDE.$ASIC.$Ch"
			error $m
		    }
		    
		    set z [catch {SpectrumPutTitle DT.$SIDE.$ASIC.$Ch [set Me]_DT.$SIDE.$ASIC.$Ch} m]

                    incr Chs $Options(R3B.DTchannels)
		    
		}
	    }
	} }
    }
}

proc map_R3B_spectra {} {
    global env
    upvar #0 SpectraHandle f
    global Options
    global NumChans
    global SIDEList
    global ASICList0
    global ASICList1
    
    set Me $::DataAcquisition::ACQSERVER
    
    set SpectraDefs  [file join $env(MIDASTempDir) SpectraDefs.dat]
    
    set z [catch {open $SpectraDefs w} f]
    if {$z != 0} then {
	puts "Cannot open file $SpectraDefs $f"
	return
    }
    
    map_spectrum "Stat 0 0 $Options(Stat.offset) $Options(Stat.shift)" 

    if {$Options(R3B_Histogramming_Enable) == 0} {close $f; return}

   set Chs 0
    
    foreach SIDE $SIDEList {
	foreach ASIC [set ASICList$SIDE] {

        if {$Options(R3B_Hist_[set SIDE].[set ASIC]_Enable) == 1} {

           if {$Chs > $Options(Hist_MaxChans)} {continue}
	    
	    puts "Mapping spectra for  SIDE $SIDE ASIC $ASIC"
	    
	    for {set Ch 0} {$Ch < $NumChans} {incr Ch} {
		
		if {$Options(R3B_Hist_E_Enable) == 1} {

                   incr Chs $Options(R3B.channels)

		   if {$Options(R3B_Hist_Hit0_Enable) == 1} {
		    
		       set group [expr $Options(R3B_GroupBase)+32+($SIDE*16)+$ASIC]
		       set item $Ch
		       map_spectrum "E.$SIDE.$ASIC.$Ch $group $item $Options(R3B.offset) $Options(R3B.shift)"
  
                   }
		   if {$Options(R3B_Hist_Hit1_Enable) == 1} {
		    
		       set group [expr $Options(R3B_GroupBase)+($SIDE*16)+$ASIC]
		       set item $Ch
		       map_spectrum "E.$SIDE.$ASIC.$Ch $group $item $Options(R3B.offset) $Options(R3B.shift)"
  
                   }
		   if {$Options(R3B_Hist_Hit01_Enable) == 1} {
		    
		       set group [expr $Options(R3B_GroupBase)+64+($SIDE*16)+$ASIC]
		       set item $Ch
		       map_spectrum "E.$SIDE.$ASIC.$Ch $group $item $Options(R3B.offset) $Options(R3B.shift)"
                   }
		}

	    }

	    for {set Ch 0} {$Ch < $NumChans} {incr Ch} {
		
		if {$Options(R3B_Hist_DT_Enable) == 1} {
		    
		    set group [expr $Options(R3B_GroupBase)+96+($SIDE*16)+$ASIC]
		    set item $Ch
		    map_spectrum "DT.$SIDE.$ASIC.$Ch $group $item $Options(R3B.DToffset) $Options(R3B.DTshift)"

                    incr Chs $Options(R3B.DTchannels)
		}

	    }
	} }
    }
    
    close $f
}

proc Setup_Asics16 {} {
       global SIDEList
       global ASICList0
       global ASICList1

#  Val input not active (permanently enabled)
#     side=0 electrons mode (N);  side=1 holes mode (P)

#    note clocks now 200 MHz

    puts "set up for 16 ASICs per side"

    foreach side $SIDEList {
    if {$side == 0} {set charge 1} else {set charge 0}
    foreach asic [set ASICList$side] {
	
       set i 9  ;#   essential for ASIC V3
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0

       set i 1  ;#   set charge mode
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  [expr 0x30 + ($charge << 7)]

#    set Reg2 bits for settings needed by readout chain

       switch $asic {
           0 -
           1 -
           2 -
           3 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x24
           }

           4 -
           5 -
           6 -
           7 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x04
           }

           8 -
           9 -
           10 -
           11 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x04
           }

           12 -
           13 -
           14 -
           15 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x54
           }

           default {
           }
        }

#    set Reg0 for LVDs settings

       switch $asic {
           0 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0xa0
           }
           1 -
           2 -
           3 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x20
           }

           4 -
           5 -
           6 -
           7 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x00
           }

           8 -
           9 -
           10 -
           11 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x00
           }

           12 -
           13 -
           14  {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x10
           }

           15  {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x50
           }

           default {
           }
        }

     }
     }

     puts "done"
}

proc Setup_Asics12 {} {
       global SIDEList
       global ASICList0
       global ASICList1

#  Val input not active (permanently enabled)
#     side=0 electrons mode (N);  side=1 holes mode (P)

#    note clocks now 200 MHz

    puts "set up for 12 ASICs per side"

    foreach side $SIDEList {
    if {$side == 0} {set charge 1} else {set charge 0}
    foreach asic [set ASICList$side] {

       set i 9  ;#   essential for ASIC V3
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0

       set i 1  ;#   set charge mode
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  [expr 0x30 + ($charge << 7)]

#    set Reg2 bits for settings needed by readout chain

       switch $asic {
           0 -
           1 -
           2 -
           3 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x24
           }

           4 -
           5 -
           6 -
           7 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x04
           }

           8 -
           9 -
           10 -
           11 {
               set i 2
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x54
           }

           default {
           }
        }

#    set Reg0 for LVDs settings

       switch $asic {
           0 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0xa0
           }
           1 -
           2 -
           3 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x20
           }

           4 -
           5 -
           6 -
           7 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x00
           }

           8 -
           9 -
           10 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x10
           }

           11 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0x50
           }

           default {
           }
        }

     }
     }

     puts "done"
}

proc Setup_Asics6 {} {
       global SIDEList
       global ASICList0
       global ASICList1

#  Val input not active (permanently enabled)

#     charge=1 electrons mode (N);  charge=0 holes mode (P)

#    note clocks now 200 MHz

    puts "set up for 6 ASICs per side"

    set side 0
    foreach asic [set ASICList$side] {

       set i 9  ;#   essential for ASIC V3
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0

#    set Reg0 for LVDs settings

       switch $asic {
           1 -
           2 -
           3 -
           4 -
           5 -
           6 {
               set i 0
               WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  0xf0
           }
        }


       switch $asic {
          1:   set charge 1
          2:   set charge 0
          3:   set charge 1
          4:   set charge 1
          5:   set charge 0
          6:   set charge 1
       }

       set i 1  ;#   set charge mode
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]  [expr 0x30 + ($charge << 7)]

       set i 2
       WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $0x74
    }

    puts "done"
}

proc Setup_Asics {} {
    global NumASICSperSIDE

    if {$NumASICSperSIDE == 16} {Setup_Asics16}
    if {$NumASICSperSIDE == 12} {Setup_Asics12}
    if {$NumASICSperSIDE == 6}  {Setup_Asics6}
}

   proc setModID {} {
       global ModID0 ModID1
       global SIDEList

     set ::R3B::R3B4K "undefined"    ;#   set to initial state in case of detector power cycle
     WriteR3BRegister 0x002 0x0000
     selectmodule [exec uname -n]    ;#   this will cause ModIDs to be discovered

     if {$SIDEList == {}} {puts "WARNING!!! no ASICS discovered"; return}

     puts "ModID side 0 = $ModID0: ModID side 1 = $ModID1"

       set side [first $SIDEList]
       WriteNetVar NetVar.R3B.ModID [set ModID$side]  ;#    use ModID to set a netvar which is available to the data acq software at GO 

       set vl [expr ($ModID1 << 8) + $ModID0]
       WriteR3BRegister 0x002  $vl

   }

   proc do_Calibrate {} {
       global SIDEList
       global ASICList0
       global ASICList1	

	puts "entering do_Calibrate"

          foreach side $SIDEList {
            foreach asic [set ASICList$side] {

              set i 5    ;#  REG5
              set REG5.$side.$asic [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]   ;#  save current value
              WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] 0   ;# disable data output from asic

            }
          }
   
       run_calibration

          foreach side $SIDEList {
            foreach asic [set ASICList$side] {

              set i 5    ;#  REG5
              WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] [set REG5.$side.$asic]   ;#  restore setting

            }
          }
	puts "exiting do_Calibrate"
   }


proc run_calibration {} {

       InsertLog "running calibrate for $::R3B::R3B4K"

       set BAK [ReadR3BRegister 0x200]

       WriteR3BRegister 0x200 0x00ff0000   ;#  soft reset active
        after 100
       WriteR3BRegister 0x200 0x00000000   ;#  soft reset inactive
        after 100

       WriteR3BRegister 0x200 0x000000ff   ;#  DataIn select start
        after 100

        set Opt800.1 [ReadR3BRegister 0x201]   ;#  status 

       WriteR3BRegister 0x200 0x00000000   ;#  DataIn select stop

        InsertLog "Calibration complete for $::R3B::R3B4K: Align Done Status = [set Opt800.1]"

         set p1 [expr ([set Opt800.1] & 255)]
         set p2 [expr (([set Opt800.1] >> 8) & 255)]

         if {[expr ($p1 ^ $p2)] != 255} {
             InsertLog "Calibration failed"
         } else {
             InsertLog "Calibration OK"
         }

       WriteR3BRegister 0x200 $BAK

       InsertLog "end of calibrate for $::R3B::R3B4K"
}

   proc do_ResetTimeStamp {} {
       global Options
       global SIDEList
       global ASICList0
       global ASICList1

	 
           puts "entering do_ResetTimeStamp" 	

          set reg [ReadR3BRegister 0x300]
          WriteR3BRegister 0x300  [expr ($reg & 0xfffffffe)]    ;# deassert Val_trig (bit 0)

          set reg [ReadR3BRegister 0x400]
          WriteR3BRegister 0x400  [expr ($reg & 0x0000ffff) | (($Options(R3B_Fpga_ts_reg) & 0x0000ffff) << 16)]    ;# set Fpga_ts_reg

          set reg [ReadR3BRegister 0x400]
          WriteR3BRegister 0x400  [expr ($reg & 0xfffff0fff) | (($Options(R3B_Asic_ts_PulseWidth) & 15) << 12)]    ;# set Asic_ts_PulseWidth

	 foreach side $SIDEList {
             foreach asic [set ASICList$side] {

		      set i 1    ;#  REG1
		      set reg [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
		      set vl [expr ($reg & 0xcf) | 0x30]                                             ;# VAL input not active
		      WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
		}
	}

        
          set reg [ReadR3BRegister 0x400]
          WriteR3BRegister 0x400  [expr ($reg & 0xffffffcf) | 0x10]    ;# connect ts_rst to asic_val


	foreach side $SIDEList {
             foreach asic [set ASICList$side] {

		      set i 1    ;#  REG1
		      set reg [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
		      set vl [expr ($reg & 0xcf) | 0x20]                                             ;# Time Stamp Reset -> VAL input
		      WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
		}
	}



          set reg [ReadR3BRegister 0x400]
          WriteR3BRegister 0x400  [expr ($reg & 0xffffffcf)]    ;# connect cal_val to ASIC_val


	 foreach side $SIDEList {
             foreach asic [set ASICList$side] {

		      set i 1    ;#  REG1
		      set reg [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
		      set vl [expr ($reg & 0xcf) | 0x30]                                             ;# VAL input not active
		      WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
		}
	}
          
          
          InsertLog "Time Stamp Reset complete  DCM status [ReadR3BRegister [expr 0x400 + 1]]"

   }

   proc do_SetExternalClock {} {
       global env

#    set clock source via CBS to external

          switch [exec /bin/uname -p] {
             ppc {
                set z [catch {exec [file join $env(MIDASBASE) R3B CBS cbs.exe] -x -c} m]
                if {$z != 0} {puts "$m"}              
             }
             default {
                 puts "not available for this processor"
             }
          }

#    PLL disable   (to be confirmed)

#    Clk_sel to external 

          set reg0 [ReadR3BRegister 0x400]
          set reg0 [expr $reg0 | 2]
          WriteR3BRegister 0x400 $reg0

#    reset custom DCM

          set reg0 [ReadR3BRegister 0x400]
          set reg0 [expr $reg0 | 1]   ;#  reset
          WriteR3BRegister 0x400 $reg0
          set reg0 [expr $reg0 & 0x7ffe]    ;#   not reset
          WriteR3BRegister 0x400 $reg0

          InsertLog "Clock source now set to external  DCM status [ReadR3BRegister 0x401] "

   }

   proc do_SetInternalClock {} {
       global env

#    set clock source via CBS to internal

          switch [exec /bin/uname -p] {
              ppc {
                 set z [catch {exec [file join $env(MIDASBASE) R3B CBS cbs.exe] -i -c} m]
                 if {$z != 0} {puts "$m"}
              }
              default {
                 puts "not available for this processor"
              }
          }

#    PLL enable

          switch [exec /bin/uname -p] {
              ppc {
                 set z [catch {exec [file join $env(MIDASBASE) R3B PLL pll.exe] -s 0xe8 0x00 0x00 0x0a 0x20 0x01} m]
                 if {$z != 0} {puts "$m"}
              }
              default {
                 puts "not available for this processor"
              }
          }

#    Clk_sel to internal 

          set reg0 [ReadR3BRegister 0x400]
          set reg0 [expr $reg0 & 0x7ffd]
          WriteR3BRegister 0x400 $reg0]

#    reset custom DCM

          set reg0 [ReadR3BRegister 0x400]
          set reg0 [expr $reg0 | 1]   ;#  reset
          WriteR3BRegister 0x400 $reg0
          set reg0 [expr $reg0 & 0x7ffe]    ;#   not reset
          WriteR3BRegister 0x400 $reg0

          InsertLog "Clock source now set to internal  DCM status [ReadR3BRegister 0x401] "
     }


proc doSetUp {} {
    global Options
    
    puts "executing doSetUp"
    
    Set_Options; make-UIregisters

    set Me $::DataAcquisition::ACQSERVER
    
    DataAcqCommand 1 "Halt"

     setModID
    
    unmap_spectra
    delete_spectra
    
    CreateSpectrum1D Stat $Options(Stat.channels)
    set z [catch {SpectrumPutTitle Stat [set Me]_Stat} m]
    
    CreateSpectrum1D Rate $Options(Rate.channels)
    set z [catch {SpectrumPutTitle Rate [set Me]_Rate} m]
    
    create_R3B_spectra
    
    map_R3B_spectra 
    
    map_spectra

    ParameterRestore $Options(ASIC.settings) -1

    Setup_Electronics   ;# do AFTER restore which allows force of certain settings - NOTE finishes with an ASIC Reset
    
    DataAcqCommand 3 "Setup"
    
    WriteNetVar NetVar.DACQ.PollTicks 100
    WriteNetVar NetVar.DACQ.Crdoutenable 1   ;# enable data readout
    
    puts "completed doSetUp"

}

proc doReset {} {

     puts "executing doReset"

    DataAcqCommand 1 "Halt"

     Reset_Electronics 

     puts "completed doReset"
}

proc doGo {} {

puts "executing doGo"

     Set_Options; make-UIregisters

     if {[ReadNetVar NetVar.DACQ.Ctxenable] == 1}  {
         DataAcqCommand 3 "connect xfer stream #0" 0 2
     }

#     Reset_Electronics   ;#   removed from GO for R3B because it causes an ASICReset  which upsets the readout calibration 

     DataAcqCommand 2 "Go"

     Go_Electronics

puts "completed  doGo"

}

proc Reset_Electronics {} {    ;#   used by RESET

     WriteR3BRegister 0x100 2        ;#  set reset DMA
     WriteR3BRegister 0x100 0        ;#  and clear

     WriteR3BRegister 0x106 0        ;#  and clear rd_ptr

     WriteR3BRegister 0x200 0x00000000   ;#  initialize control register

     do_ASICReset     ;#  generate RESET ASIC pulse   -   Note after this a  readout calibration is required 
}

   proc do_ASICReset {} {
       global Options

       puts "entering do_ASICReset"

       WriteR3BRegister 0x304 $Options(R3B_ResetPulseWidth)
       WriteR3BRegister 0x300 0x10   ;#  reset to load values
       WriteR3BRegister 0x300 0

       WriteR3BRegister 0x300 8; WriteR3BRegister 0x300 0;#  generate reset pulse
       after 10   
   }

proc Setup_Electronics {} {    ;#   used by SETUP
   global Options

   puts "starting Setup_Electronics"

    if {$Options(R3B_ClockSource) == "E"} {do_SetExternalClock} else {do_SetInternalClock}

    set Me $::DataAcquisition::ACQSERVER

     WriteR3BRegister 0x100 2        ;#  set reset DMA
     WriteR3BRegister 0x100 0        ;#  and clear

     WriteR3BRegister 0x102 0x0f000000   ;# set DMA MEM Base
     WriteR3BRegister 0x103 2048         ;# set DMA burst length
     WriteR3BRegister 0x105 8192  ;# set DMA MEM Length (units of burst length) (0x01000000/2048) == (16M/2K)
     WriteR3BRegister 0x104 256         ;# set DMA Buffer Length (units of burst length) (32 buffers)  (8192/32)

     WriteR3BRegister 0x106 0        ;#  and clear rd_ptr

     Setup_Asics   ;#  put R2 to default settings

     WriteR3BRegister 0x301 $Options(R3B_ValPulseWidth)     ;# Validation Pulse Width 
     WriteR3BRegister 0x302 $Options(R3B_ValPulseDelay)     ;# Validation Pulse Delay
     WriteR3BRegister 0x304 $Options(R3B_ResetPulseWidth)   ;# Reset Pulse Width   

     #do_ASICReset    

     do_ResetTimeStamp 

     do_Calibrate    ;#   this must be repeated after any ASICReset

   puts "finishing Setup_Electronics  DCM status [ReadR3BRegister [expr 0x400 + 1]]"

}

proc Stop_Electronics {} {

     WriteR3BRegister 0x200 0x00000000   ;#  packet disable

     WriteR3BRegister 0x100 0        ;#    unset enable DMA

}

proc Go_Electronics {} {

    set Me $::DataAcquisition::ACQSERVER

     WriteR3BRegister 0x100 2        ;#  set reset DMA
     WriteR3BRegister 0x100 0        ;#  and clear

     WriteR3BRegister 0x106 0        ;#  and clear rd_ptr

     WriteR3BRegister 0x100 1        ;#  enable  DMA

     WriteR3BRegister 0x200 0x0000ff00   ;#  packet enable

}


   proc ParameterRestore {DBkey p} {
       global env
       global NumASICs
       global NumASICSperSIDE
       global SIDEList
       global ASICList0
       global ASICList1
       global ModID0 ModID1

      set ::R3B::DBNodeBase $DBkey
      set ::R3B::DBNodeName $ModID0
      
      puts "[date +%Y%h%d-%H.%M.%S] Restoring settings from $::R3B::DBRootName/$::R3B::DBNodeBase/$::R3B::DBNodeName  with selection $p"

#    first restore ASIC settings

   foreach side $SIDEList {
     if {$side == ""} {continue}
     foreach asic [set ASICList$side] {
         if {$asic == ""} {continue}

         set ASIC [expr ($side*16) + $asic]

         set s [expr (($p >> ($side * 16)) >> ($NumASICSperSIDE - $asic - 1)) & 1]
         if {$s == 1} {
         
	     for {set i 0} {$i <= 5} {incr i} {     ;# allow REG0 since 22 Jul 2015 (ASIC  V3)
              set vl [DBGetValue ASIC.$ASIC.$i]
              if {$i == 1} {set vl [expr ($vl & 0xc0) + 0x30]}
              WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
            }

            for {set i 10} {$i <= 21} {incr i} {
              set vl [DBGetValue ASIC.$ASIC.$i]
              WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
            }

#           for {set i 64} {$i <= 79} {incr i} {
#             set vl [DBGetValue ASIC.$ASIC.$i]
#             WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
#           }

#           for {set i 80} {$i <= 95} {incr i} {
#             set vl [DBGetValue ASIC.$ASIC.$i]
#             WriteR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i] $vl
#           }
         }

      }
   }

      puts "[date +%Y%h%d-%H.%M.%S] Settings restored from $::R3B::DBRootName/$::R3B::DBNodeBase/$::R3B::DBNodeName"


#       ::R3B::WriteRegister HasBeenRestored 1

#        Elog INFORMATION "FEE64 settings restored from $::R3B::DBRootName/$::R3B::DBNodeBase for $::DataAcquisition::ACQSERVER"

   }

   proc ParameterSave {DBkey p} {
       global env
       global NumASICs
       global NumASICSperSIDE
       global SIDEList
       global ASICList0
       global ASICList1
       global ModID0 ModID1

      set ::R3B::DBNodeBase $DBkey
      set ::R3B::DBNodeName $ModID0
      
      puts "[date +%Y%h%d-%H.%M.%S] Saving settings to $::R3B::DBRootName/$::R3B::DBNodeBase/$::R3B::DBNodeName  with selection $p"

#    first save ASIC settings

   foreach side $SIDEList {
     if {$side == ""} {continue}
     foreach asic [set ASICList$side] {
         if {$asic == ""} {continue}

         set ASIC [expr ($side*16) + $asic]

         set s [expr (($p >> ($side * 16)) >> ($NumASICSperSIDE - $asic - 1)) & 1]
         if {$s == 1} {

            for {set i 0} {$i <= 5} {incr i} {   ;# allow REG0 since 22 Jul 2015 (ASIC  V3);  Note regs 6 & 7 are read only
              set vl [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
              DBSetValue ASIC.$ASIC.$i [format "0x%02x" [expr $vl & 255]]
            }

            for {set i 10} {$i <= 21} {incr i} {
              set vl [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
              DBSetValue ASIC.$ASIC.$i [format "0x%02x" [expr $vl & 255]]
            }

#           for {set i 64} {$i <= 79} {incr i} {
#             set vl [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
#             DBSetValue ASIC.$ASIC.$i [format "0x%02x" [expr $vl & 255]]
#           }

#           for {set i 80} {$i <= 95} {incr i} {
#             set vl [ReadR3BASICRegister [expr ($side << 18) + ($asic << 10) + $i]]
#             DBSetValue ASIC.$ASIC.$i [format "0x%02x" [expr $vl & 255]]
#           }
         }
      }
   }

      puts "[date +%Y%h%d-%H.%M.%S] Settings saved to $::R3B::DBRootName/$::R3B::DBNodeBase/$::R3B::DBNodeName"

#       ::R3B::WriteRegister HasBeenRestored 1

#        Elog INFORMATION "R3B4K settings saved to $::R3B::DBRootName/$::R3B::DBNodeBase for $::DataAcquisition::ACQSERVER"

   }

    proc HistogramZeroAll {args} {

       set SpecNames [SpectrumNames]
       set NumSpec 0

       ZeroSpectrum Stat
       incr NumSpec

      foreach item $SpecNames {
         if {[first $item] != 1} {continue}   ;#   not a 1D histogram !!

         set SpecName [second $item]

         ZeroSpectrum $SpecName
         incr NumSpec
      }

       puts "$NumSpec spectra zeroed"
    }

    proc HistogramZero {p} {   ;#  supports Zero for histograms in selected ASICs
       global env
       global NumASICs
       global NumASICSperSIDE
       global SIDEList
       global ASICList0
       global ASICList1

       set SpecNames [SpectrumNames]
       set NumSpec 0

       ZeroSpectrum Stat
       incr NumSpec

   foreach side $SIDEList {
     foreach asic [set ASICList$side] {

         set ASIC [expr ($side*$NumASICSperSIDE) + $asic]

         if {[expr ($p >> [expr $NumASICs-$ASIC-1]) & 1] == 1} {

            foreach item $SpecNames {
               if {[first $item] != 1} {continue}   ;#   not a 1D histogram !!

               set SpecName [second $item]

               if {$side != [second [split $SpecName .]]} {continue}
               if {$asic != [third [split $SpecName .]]} {continue}

               ZeroSpectrum $SpecName
               incr NumSpec
            }
         }

      }
   }

       puts "$NumSpec spectra zeroed"
    }


 proc SpecSave1D {spec base range type title counts} {
    variable SpecPath
     
     set f3 [file join $SpecPath $spec.msf]
     set z1 [catch {open $f3 w} m1]
     set z2 [catch {file attributes $f3 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f3 - $m1"
      }
      if {$z2 != 0} then {
          puts "error accessing $f3 - $m2"
      }

      set f $m1

      fconfigure $f -translation binary -encoding binary

  set ctime [clock format [clock seconds] -format {%d-%b-%Y %H:%M:%S}]
  set mtime [clock format [clock seconds] -format {%d-%b-%Y %H:%M:%S}]

      puts -nonewline $f [binary format i1 412900921]   ;#   magic number
      puts -nonewline $f [binary format i1 1]           ;#   version
      puts -nonewline $f [binary format a32 $spec]      ;#   spectrum name
      puts -nonewline $f [binary format i1 1]           ;#   dimension
      puts -nonewline $f [binary format a20 $ctime]     ;#   creation time/date
      puts -nonewline $f [binary format a20 $mtime]     ;#   modification time/date
      puts -nonewline $f [binary format i1 $base]       ;#   base0

      for {set i 2} {$i <=8} {incr i} {
         puts -nonewline $f [binary format i1 -1]     ;#   baseI
      }

      puts -nonewline $f [binary format i1 $range]    ;#   range0

      for {set i 2} {$i <=8} {incr i} {
         puts -nonewline $f [binary format i1 -1]     ;#   rangeI
      }

      puts -nonewline $f [binary format i1 436]        ;#   spectrum title offset  (was -1)

      for {set i 2} {$i <=32} {incr i} {
         puts -nonewline $f [binary format i1 -1]     ;#  information pointers
      }

      for {set i 1} {$i <=8} {incr i} {
         puts -nonewline $f [binary format i1 -1]     ;#  annotation pointers
      }

      for {set i 1} {$i <=8} {incr i} {
         puts -nonewline $f [binary format i1 -1]     ;#  calibration pointers
      }

      for {set i 1} {$i <=8} {incr i} {
         puts -nonewline $f [binary format i1 -1]     ;#  efficiency pointers
      }

                                                     ;#  data array 1 descriptor 
      puts -nonewline $f [binary format i1 0]        ;#  array layout
      puts -nonewline $f [binary format i1 $type]    ;#  array type  (32 bit unsigned)
      puts -nonewline $f [binary format i1 0]        ;#  reserved 
      puts -nonewline $f [binary format i1 0]        ;#  reserved
      puts -nonewline $f [binary format i1 0]        ;#  spectrum counts offset

                                                     ;#  data array 2 descriptor 
      puts -nonewline $f [binary format i1 0]        ;#  array layout
      puts -nonewline $f [binary format i1 $type]    ;#  array type 
      puts -nonewline $f [binary format i1 0]        ;#  reserved 
      puts -nonewline $f [binary format i1 0]        ;#  reserved
      puts -nonewline $f [binary format i1 -1]       ;#  spectrum counts offset

      puts -nonewline $f [binary format i1 436]      ;#  base of string space
      puts -nonewline $f [binary format i1 20]       ;#  string free space
      puts -nonewline $f [binary format i1 76]       ;#  top of string space

      puts -nonewline $f [binary format i1 512]      ;#  base of counts space
      puts -nonewline $f [binary format i1 $range]   ;#  counts free space
      puts -nonewline $f [binary format i1 $range]   ;#  top of counts space
 
      puts -nonewline $f [binary format a20 $title]  ;# string space - title

      for {set i 456} {$i < 512} {incr i} {          ;#
         puts -nonewline $f [binary format c1 0]     ;#  unused
      }

      puts -nonewline $f [binary format i* $counts]  ;# counts

      close $f
}

    proc HistogramSaveAll {DBkey p} {
       global env
       global NumASICs
       global NumASICSperSIDE
       global SIDEList
       global ASICList0
       global ASICList1
       variable SpecPath

      set f1 $DBkey
      set z1 [catch {file mkdir $f1} m1]
      set z2 [catch {file attributes $f1 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f1 - $m1"
      }
      if {$z2 != 0} then {
          puts "error accessing $f1 - $m2"
      }

      set f1 [file join $f1 $env(ComputerName)]
      set z1 [catch {file mkdir $f1} m1]
      set z2 [catch {file attributes $f1 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f1 - $m1"
      }
      if {$z2 != 0} then {
          puts "error accessing $f1 - $m2"
      }

      set f2 [file join $f1 [date +%Y%h%d-%H.%M.%S]]
      set z1 [catch {file mkdir $f2} m1]
      set z2 [catch {file attributes $f2 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f2 - $m1"
      }
      if {$z2 != 0} then {
          puts "error accessing $f2 - $m2"
      }

      puts "[date +%Y%h%d-%H.%M.%S] Saving Histograms to $f2 for $::DataAcquisition::ACQSERVER"

       set SpecPath $f2

       set SpecNames [SpectrumNames]
       set NumSpec 0

         set SpecName Stat

         set info [SpectrumDetails $SpecName]
         set channels [lindex $info 2]

         set title [SpectrumGetTitle $SpecName]
         set counts [SpectrumRead1D $SpecName 0 $channels]

         set z [catch {SpecSave1D $SpecName 0 $channels 5 $title $counts} m]
         if {$z!=0} then {
             puts "Error attempting to save spectrum $SpecName: $m"
         } else {
             incr NumSpec
         }

       foreach item $SpecNames {
         if {[first $item] != 1} {continue}    ;#  not a 1D histogram

         set SpecName [second $item]

         set info [SpectrumDetails $SpecName]
         set channels [lindex $info 2]

         set title [SpectrumGetTitle $SpecName]
         set counts [SpectrumRead1D $SpecName 0 $channels]

         set z [catch {SpecSave1D $SpecName 0 $channels 5 $title $counts} m]
         if {$z!=0} then {
             puts "Error attempting to save spectrum $SpecName: $m"
         } else {
             incr NumSpec
         }
       }

      puts "[date +%Y%h%d-%H.%M.%S] $NumSpec Histograms saved for $::DataAcquisition::ACQSERVER"

    }

    proc HistogramSave {DBkey p} {
       global env
       global NumASICs
       global NumASICSperSIDE
       global SIDEList
       global ASICList0
       global ASICList1
       variable SpecPath

      set f1 $DBkey
      set z1 [catch {file mkdir $f1} m1]
      set z2 [catch {file attributes $f1 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f1 - $m1"
          return
      }
      if {$z2 != 0} then {
          puts "error accessing $f1 - $m2"
      }

      set f1 [file join $f1 $env(ComputerName)]
      set z1 [catch {file mkdir $f1} m1]
      set z2 [catch {file attributes $f1 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f1 - $m1"
          return
      }
      if {$z2 != 0} then {
          puts "error accessing $f1 - $m2"
      }

      set f2 [file join $f1 [date +%Y%h%d-%H.%M.%S]]
      set z1 [catch {file mkdir $f2} m1]
      set z2 [catch {file attributes $f2 -permissions rwxrwxrwx} m2]

      if {$z1 != 0} then {
          puts "Cannot create $f2 - $m1"
          return
      }
      if {$z2 != 0} then {
          puts "error accessing $f2 - $m2"
      }

      puts "[date +%Y%h%d-%H.%M.%S] Saving Histograms to $f2 for $::DataAcquisition::ACQSERVER with selection $p"

       set SpecPath $f2

       set SpecNames [SpectrumNames]
       set NumSpec 0

         set SpecName Stat

         set info [SpectrumDetails $SpecName]
         set channels [lindex $info 2]

         set title [SpectrumGetTitle $SpecName]
         set counts [SpectrumRead1D $SpecName 0 $channels]

         set z [catch {SpecSave1D $SpecName 0 $channels 5 $title $counts} m]
         if {$z!=0} then {
             puts "Error attempting to save spectrum $SpecName: $m"
         } else {
             incr NumSpec
         }

   foreach side $SIDEList {
     if {$side == ""} {continue}
     foreach asic [set ASICList$side] {
         if {$asic == ""} {continue}

         set s [expr (($p >> ($side * 16)) >> ($NumASICSperSIDE - $asic - 1)) & 1]
         if {$s == 1} {

       foreach item $SpecNames {
         if {[first $item] != 1} {continue}    ;#  not a 1D histogram

         set SpecName [second $item]

         if {$side != [second [split $SpecName .]]} {continue}
         if {$asic != [third [split $SpecName .]]} {continue}

         set info [SpectrumDetails $SpecName]
         set channels [lindex $info 2]

         set title [SpectrumGetTitle $SpecName]
         set counts [SpectrumRead1D $SpecName 0 $channels]

         set z [catch {SpecSave1D $SpecName 0 $channels 5 $title $counts} m]
         if {$z!=0} then {
             puts "Error attempting to save spectrum $SpecName: $m"
         } else {
             incr NumSpec
         }

       }
      }
     }
   }

      puts "[date +%Y%h%d-%H.%M.%S] $NumSpec Histograms saved for $::DataAcquisition::ACQSERVER"

    }

    proc TerminateExec {} {

        unmap_spectra
        DataAcqCommand 22 "Terminating DataAcq program" 0 2

    }

    proc LoadExec {} {
      global Options
      global env

       set f [file join $env(MIDASBASE) Data_Acq bin linux-ppc_4xx $Options(DataAcqPgm)]

       set z [catch {eval exec nice $f &} m]
       if {$z != 0} {puts "RunControl LoadExec from $f returned z=$z and $m"}

    }

    proc FlashPgm {} {
       global env

       set f [file join $env(MIDASBASE) R3B FlashPgm.csh]

       set z [catch {exec $f > /tmp/FlashPgm.log} m]

       if {$z != 0} {
		set file_name [file join / tmp FlashPgm.log]
		puts "RunControl FlashPgm from $f returned z=$z and $m"
         	set z [catch {open $file_name a} file_id]
		if { $z != 0 }  { puts "\nFile open failed $file_name :- $z" ; break }
         	puts $file_id "FAILED\nMessage :$m\n"
         	set z [catch {close $file_id } close_error]
	}

    }



    variable registerlist
    lappend registerlist [list NetVar.R3B.ModID  "-n R3B_ModID -t int"]
    lappend registerlist [list NetVar.R3B.ModuleNum  "-n R3B_ModuleNum -t int"]

    make-UIregisters

    Set_Options

  

